@@ -40,7 +40,7 @@ def isJobFinishedTerminatedOrFailed(job):
 def handle_response(response):
     data = None
     if response.status_code in [500, 501]:
-        LOGGER.warn("McritClient received status code 500 from MCRIT.")
+        LOGGER.warning("McritClient received status code 500 from MCRIT.")
     elif response.status_code in [400, 404, 410]:
         # nothing to here as of now
         pass
@@ -58,7 +58,7 @@ class McritClient:
             self.mcrit_server = mcrit_server
 
     def _getMatchingRequestParams(
-        self, minhash_threshold, pichash_size, force_recalculation
+        self, minhash_threshold, pichash_size, force_recalculation, band_matches_required
     ):
         params = {}
         if minhash_threshold is not None:
@@ -67,6 +67,8 @@ class McritClient:
             params["pichash_size"] = pichash_size
         if force_recalculation is not None:
             params["force_recalculation"] = force_recalculation
+        if band_matches_required is not None:
+            params["band_matches_required"] = band_matches_required
         return params
 
     def respawn(self):
@@ -108,6 +110,15 @@ class McritClient:
     ### Families 
     ###########################################
 
+    def modifyFamily(self, family_id, family_name=None, is_library=None):
+        update_dict = {}
+        if family_name is not None:
+            update_dict["family_name"] = family_name
+        if is_library is not None:
+            update_dict["is_library"] = is_library
+        response = requests.put(f"{self.mcrit_server}/families/{family_id}", update_dict)
+        return handle_response(response)
+
     def getFamily(self, family_id: int, with_samples=True) -> Optional[FamilyEntry]:
         query_params = "?with_samples=true" if with_samples else "?with_samples=false"
         response = requests.get(f"{self.mcrit_server}/families/{family_id}{query_params}")
@@ -142,6 +153,19 @@ class McritClient:
             return True
         return False
 
+    def modifySample(self, sample_id, family_name=None, version=None, component=None, is_library=None):
+        update_dict = {}
+        if family_name is not None:
+            update_dict["family_name"] = family_name
+        if version is not None:
+            update_dict["version"] = version
+        if component is not None:
+            update_dict["component"] = component
+        if is_library is not None:
+            update_dict["is_library"] = is_library
+        response = requests.put(f"{self.mcrit_server}/samples/{sample_id}", update_dict)
+        return handle_response(response)
+
     def deleteSample(self, sample_id):
         response = requests.delete(f"{self.mcrit_server}/samples/{sample_id}")
         return handle_response(response)
@@ -202,6 +226,10 @@ class McritClient:
         if data is not None:
             return FunctionEntry.fromDict(data)
 
+    def completeMinhashes(self):
+        response = requests.get(f"{self.mcrit_server}/complete_minhashes")
+        return handle_response(response)
+
     ###########################################
     ### Matching 
     ###########################################
@@ -211,10 +239,13 @@ class McritClient:
         smda_report: SmdaReport,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         force_recalculation=False,
     ) -> str:
         smda_json = smda_report.toDict()
-        params = self._getMatchingRequestParams(minhash_threshold, pichash_size, force_recalculation)
+        params = self._getMatchingRequestParams(
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
+        )
         response = requests.post(f"{self.mcrit_server}/query", json=smda_json, params=params)
         return handle_response(response)
 
@@ -224,6 +255,7 @@ class McritClient:
         base_address: int,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         disassemble_locally=True,
         force_recalculation=False,
     ) -> str:
@@ -236,10 +268,13 @@ class McritClient:
                 smda_report,
                 minhash_threshold=minhash_threshold,
                 pichash_size=pichash_size,
+                band_matches_required=band_matches_required,
                 force_recalculation=force_recalculation,
             )
 
-        params = self._getMatchingRequestParams(minhash_threshold, pichash_size, force_recalculation)
+        params = self._getMatchingRequestParams(
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
+        )
         response = requests.post(f"{self.mcrit_server}/query/binary/mapped/{base_address}", binary, params=params)
         return handle_response(response)
 
@@ -248,6 +283,7 @@ class McritClient:
         binary: bytes,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         disassemble_locally=True,
         force_recalculation=False,
     ) -> str:
@@ -260,10 +296,13 @@ class McritClient:
                 smda_report,
                 minhash_threshold=minhash_threshold,
                 pichash_size=pichash_size,
+                band_matches_required=band_matches_required,
                 force_recalculation=force_recalculation,
             )
 
-        params = self._getMatchingRequestParams(minhash_threshold, pichash_size, force_recalculation)
+        params = self._getMatchingRequestParams(
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
+        )
 
         response = requests.post(f"{self.mcrit_server}/query/binary", binary, params=params)
         return handle_response(response)
@@ -273,10 +312,11 @@ class McritClient:
         sample_id,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         force_recalculation=False,
     ) -> None:
         params = self._getMatchingRequestParams(
-            minhash_threshold, pichash_size, force_recalculation
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
         )
         response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}", params=params)
         return handle_response(response)
@@ -287,10 +327,11 @@ class McritClient:
         other_sample_id,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         force_recalculation=False,
     ) -> str:
         params = self._getMatchingRequestParams(
-            minhash_threshold, pichash_size, force_recalculation
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
         )
         response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}/{other_sample_id}",params=params)
         return handle_response(response)
@@ -300,10 +341,11 @@ class McritClient:
         sample_ids,
         minhash_threshold=None,
         pichash_size=None,
+        band_matches_required=None,
         force_recalculation=False,
     ) -> None:
         params = self._getMatchingRequestParams(
-            minhash_threshold, pichash_size, force_recalculation
+            minhash_threshold, pichash_size, force_recalculation, band_matches_required
         )
         response = requests.get(
             f"{self.mcrit_server}/matches/sample/cross/{','.join([str(id) for id in sample_ids])}",
@@ -311,6 +353,14 @@ class McritClient:
         )
         return handle_response(response)
 
+    def getMatchFunctionVs(
+        self,
+        function_id_a:int,
+        function_id_b:int
+    ) -> None:
+        response = requests.get(f"{self.mcrit_server}/matches/function/{function_id_a}/{function_id_b}")
+        return handle_response(response)
+
     def getMatchesForPicHash(self, pichash, summary=False):
         summary_string = "/summary" if summary else ""
         response = requests.get(f"{self.mcrit_server}/query/pichash/{pichash:016x}{summary_string}")
@@ -336,6 +386,13 @@ class McritClient:
         response = requests.get(f"{self.mcrit_server}/status")
         return handle_response(response)
 
+    def getVersion(self):
+        response = requests.get(f"{self.mcrit_server}/version")
+        data = handle_response(response)
+        if isinstance(data, dict) and "version" in data:
+            return data["version"]
+        return None
+
     def getJobCount(self, filter=None):
         query_string = ""
         if isinstance(filter, str) and filter is not None:
@@ -428,10 +485,10 @@ class McritClient:
         return handle_response(response)
 
     ###########################################
-    ### Import / Export
+    ### Unique Blocks
     ###########################################
 
-    def getUniqueBlocksForSamples(self, sample_ids: List[int]) -> Dict:
+    def requestUniqueBlocksForSamples(self, sample_ids: List[int]) -> Dict:
         if isinstance(sample_ids, list) and all(isinstance(item, int) for item in sample_ids):
             sample_ids_as_str = ",".join([str(sample_id) for sample_id in sample_ids])
             response = requests.get(f"{self.mcrit_server}/uniqueblocks/samples/{sample_ids_as_str}")
@@ -440,7 +497,7 @@ class McritClient:
             raise ValueError("sample_ids must be a list of int.")
         return result_data
 
-    def getUniqueBlocksForFamily(self, family_id: int) -> Dict:
+    def requestUniqueBlocksForFamily(self, family_id: int) -> Dict:
         if isinstance(family_id, int):
             response = requests.get(f"{self.mcrit_server}/uniqueblocks/family/{family_id}")
             result_data = handle_response(response)
