@@ -6,8 +6,6 @@ import logging
 from typing import Dict, List
 
 from smda.common.SmdaReport import SmdaReport
-from smda.common.SmdaInstruction import SmdaInstruction
-from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper
 
 from mcrit.config.McritConfig import McritConfig
 from mcrit.config.MinHashConfig import MinHashConfig
@@ -21,8 +19,10 @@ from mcrit.queue.QueueRemoteCalls import QueueRemoteCaller, NoProgressReporter
 from mcrit.storage.FamilyEntry import FamilyEntry
 from mcrit.storage.FunctionEntry import FunctionEntry
 from mcrit.storage.SampleEntry import SampleEntry
+from mcrit.storage.MatchedFunctionEntry import MatchedFunctionEntry
 from mcrit.storage.StorageFactory import StorageFactory
 from mcrit.minhash.MinHash import MinHash
+import mcrit.matchers.MatcherInterface as MatcherInterface
 from mcrit.Worker import Worker
 
 logging.basicConfig(level=logging.INFO)
@@ -62,7 +62,8 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         self._storage_config = config.STORAGE_CONFIG
         self._minhash_config = config.MINHASH_CONFIG
         self._shingler_config = config.SHINGLER_CONFIG
-        self._storage = StorageFactory.getStorage(config.STORAGE_CONFIG)
+        self._queue_config = config.QUEUE_CONFIG
+        self._storage = StorageFactory.getStorage(config)
         # config.QUEUE_CONFIG.QUEUE_METHOD = QueueFactory.QUEUE_METHOD_FAKE
         queue = QueueFactory().getQueue(config, storage=self._storage, consumer_id="index")
         self.search_query_parser = SearchQueryParser()
@@ -80,6 +81,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 "minhash_config": self._minhash_config.toDict(),
                 "shingler_config": self._shingler_config.toDict(),
                 "storage_config": self._storage_config.toDict(),
+                "queue_config": self._queue_config.toDict(),
             },
             "stats": self._storage.getStats(),
             "storage": self._storage.getContent(),
@@ -90,8 +92,14 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         self._minhash_config = MinHashConfig.fromDict(storage_data["config"]["minhash_config"])
         self._shingler_config = ShinglerConfig.fromDict(storage_data["config"]["shingler_config"])
         self._storage_config = StorageConfig.fromDict(storage_data["config"]["storage_config"])
+        self._queue_config = StorageConfig.fromDict(storage_data["config"]["queue_config"])
+        mcrit_config = McritConfig()
+        mcrit_config.MINHASH_CONFIG = self._minhash_config
+        mcrit_config.SHINGLER_CONFIG = self._shingler_config
+        mcrit_config.STORAGE_CONFIG = self._storage_config
+        mcrit_config.QUEUE_CONFIG = self._queue_config
         # reinitialize
-        self._storage = StorageFactory.getStorage(self._storage_config)
+        self._storage = StorageFactory.getStorage(mcrit_config)
         self._storage.setContent(storage_data["storage"])
 
     def getExportData(self, sample_ids=None, compress_data=False):
@@ -232,7 +240,13 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def getMatchesForSample(self, sample_id, minhash_threshold=None):
     def getMatchesForSampleVs(self, sample_id, other_sample_id, minhash_threshold=None):
     def getAggregatedMatchesForSample(self, sample_id, minhash_threshold=None):
-    addBinarySample(self, binary, is_dump, bitness, base_address, progress_reporter=NoProgressReporter()):
+    def getUniqueBlocks(self, sample_ids):
+    def addBinarySample(self, binary, is_dump, bitness, base_address,):
+    # these need to be jobs to ensure database consistency
+    def deleteSample(self, sample_id):
+    def deleteFamily(self, family_id, keep_samples=False):
+    def modifyFamily(self, family_id, update_information):
+    def modifySample(self, sample_id, update_information):
     """
 
     #### NOT REDIRECTED ####
@@ -268,6 +282,40 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
             sample_to_job_id[id] = job_id
         return self.combineMatchesToCross(sample_to_job_id, await_jobs=[*sample_to_job_id.values()], force_recalculation=force_recalculation)
 
+    def getMatchesFunctionVs(self, function_id_a:int, function_id_b:int):
+        function_entry_a = self._storage.getFunctionById(function_id_a, with_xcfg=True)
+        function_entry_b = self._storage.getFunctionById(function_id_b, with_xcfg=True)
+        if function_entry_a is None or function_entry_b is None:
+            return
+        sample_entry_a = self._storage.getSampleById(function_entry_a.sample_id)
+        sample_entry_b = self._storage.getSampleById(function_entry_b.sample_id)
+        minhash_a = function_entry_a.getMinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
+        minhash_b = function_entry_b.getMinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
+
+        score = None
+        if minhash_a.minhash and minhash_b.minhash:
+            score = MinHash.calculateMinHashScore(
+                    minhash_a.minhash, minhash_b.minhash, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS
+                )
+        match_flags = 0
+        match_flags += MatcherInterface.IS_MINHASH_FLAG if score is not None and score >= self._minhash_config.MINHASH_MATCHING_THRESHOLD else 0
+        match_flags += MatcherInterface.IS_PICHASH_FLAG if function_entry_a.pichash == function_entry_b.pichash else 0
+        match_flags += MatcherInterface.IS_LIBRARY_FLAG if sample_entry_b.is_library else 0
+        match_tuple = [
+            function_entry_b.family_id, 
+            function_entry_b.sample_id, 
+            function_entry_b.function_id, 
+            score, 
+            match_flags]
+        result = {
+            "function_entry_a": function_entry_a.toDict(),
+            "function_entry_b": function_entry_b.toDict(),
+            "sample_entry_a": sample_entry_a.toDict(),
+            "sample_entry_b": sample_entry_b.toDict(),
+            "match_entry": MatchedFunctionEntry(function_id_a, function_entry_a.binweight, function_entry_a.offset, match_tuple).toDict()
+        }
+        return result
+
     #### SIMPLE LOOKUPS ####
     def getFamily(self, family_id):
         return self._storage.getFamily(family_id)
@@ -281,12 +329,6 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def isSampleId(self, sample_id):
         return self._storage.isSampleId(sample_id)
 
-    def deleteSample(self, sample_id):
-        return self._storage.deleteSample(sample_id)
-
-    def deleteFamily(self, family_id, keep_samples=False):
-        return self._storage.deleteFamily(family_id, keep_samples=keep_samples)
-
     def getFunctionById(self, function_id, with_xcfg=False):
         return self._storage.getFunctionById(function_id, with_xcfg=with_xcfg)
 
@@ -336,6 +378,9 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         }
         return status
 
+    def getVersion(self):
+        return {"version": self.config.VERSION}
+
     def getMatchesForPicHash(self, pichash):
         return self._storage.getMatchesForPicHash(pichash)
         
@@ -345,22 +390,6 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def getSampleBySha256(self, sample_sha256): 
         return self._storage.getSampleBySha256(sample_sha256)
 
-    def getUniqueBlocks(self, sample_ids):
-        unique_blocks = self._storage.getUniqueBlocks(sample_ids)
-        # enrich with escaped sequences
-        sample_addr_borders = {}
-        for sample_id in sample_ids:
-            sample_entry = self._storage.getSampleById(sample_id)
-            sample_addr_borders[sample_id] = {"lower": sample_entry.base_addr, "upper": sample_entry.base_addr + sample_entry.binary_size}
-        for block_hash, entry in unique_blocks.items():
-            sample_id = entry["sample_id"]
-            escaped_sequences = []
-            for instruction in entry["instructions"]:
-                smda_instruction = SmdaInstruction(instruction)
-                escaped_sequences.append(smda_instruction.getEscapedBinary(IntelInstructionEscaper, escape_intraprocedural_jumps=True, lower_addr=sample_addr_borders[sample_id]["lower"], upper_addr=sample_addr_borders[sample_id]["upper"]))
-            unique_blocks[block_hash]["escaped_sequence"] = " ".join(escaped_sequences)
-        return unique_blocks
-
     #### SEARCH ####
 
     # When performing an initial search, the cursor should be set to None.
@@ -460,7 +489,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 if self._storage.isFamilyId(term_as_int):
                     id_match = self._storage.getFamily(term_as_int).toDict()
             else:
-                LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
+                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
         except Exception:
             pass
 
@@ -496,7 +525,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 if self._storage.isFunctionId(term_as_int):
                     id_match = self._storage.getFunctionById(term_as_int).toDict()
             else:
-                LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
+                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
         except Exception:
             pass
 
@@ -535,7 +564,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 if self._storage.isSampleId(term_as_int):
                     id_match = self._storage.getSampleById(term_as_int).toDict()
             else:
-                LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
+                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
         except Exception:
             pass
 
