@@ -2,7 +2,7 @@ import datetime
 import functools
 import logging
 import math
-from collections import defaultdict
+from collections import defaultdict, Counter
 from multiprocessing import Pool, cpu_count
 from timeit import default_timer as timer
 from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple, Union
@@ -64,7 +64,7 @@ def add_duration(func):
 
 class MatcherInterface(object):
     def __init__(
-        self, worker: "Worker", minhash_threshold=None, pichash_size=None, progress_reporter=NoProgressReporter()
+        self, worker: "Worker", minhash_threshold=None, pichash_size=None, band_matches_required=None, progress_reporter=NoProgressReporter()
     ):
         # Extended by Query, VS, Sample
         self._worker: "Worker" = worker
@@ -76,6 +76,9 @@ class MatcherInterface(object):
         self._minhash_threshold = minhash_threshold
         if pichash_size is None:
             pichash_size = self._worker.config.MINHASH_CONFIG.PICHASH_SIZE
+        if band_matches_required is None:
+            band_matches_required = self._worker.config.MINHASH_CONFIG.BAND_MATCHES_REQUIRED
+        self._band_matches_required = band_matches_required
         self._pichash_size = pichash_size
         self._additional_setup()
         self._sample_to_lib_info: Dict[int, bool]
@@ -97,91 +100,129 @@ class MatcherInterface(object):
         # Every Matcher has own implementation
         raise NotImplementedError
 
-    def _createMinHashCandidateGroups(self, pichash_matches) -> Dict[int, Set[int]]:
+    def _createMinHashCandidateGroups(self, start=0, end=None) -> Dict[int, Set[int]]:
         # Query, VS, Sample
         # Sample , Query have this plain version
         # VS adds another intersection step to this version
         # find candidates based on bands
+        if end is None:
+            end = len(self._function_entries)
         candidate_groups = {}
         function_id_to_minhash = {}
-        for function_entry in self._function_entries:
+        for function_entry in self._function_entries[start:end]:
             function_id_to_minhash[function_entry.function_id] = function_entry.getMinHash(
                 minhash_bits=self._worker._minhash_config.MINHASH_SIGNATURE_BITS
             )
-        candidate_groups = self._storage.getCandidatesForMinHashes(function_id_to_minhash)
-        candidates_to_remove = {candidate_fid: set([candidate_fid]) for candidate_fid in candidate_groups}
-        if self._worker._minhash_config.PICHASH_IMPLIES_MINHASH_MATCH:
-            for pichash_match in pichash_matches:
-                if pichash_match[0] in candidates_to_remove:
-                    candidates_to_remove[pichash_match[0]].add(pichash_match[2])
-        for fid, removals in candidates_to_remove.items():
-            candidate_groups[fid].difference_update(removals)
+        candidate_groups = self._storage.getCandidatesForMinHashes(function_id_to_minhash, band_matches_required=self._band_matches_required)
+
         return candidate_groups
 
-    def _createMatchingCache(self, function_ids):
+    def _createMatchingCache(self, candidate_groups):
         # Query, VS, Sample
         # This version is used by VS, Sample
         # Own version by Query
-        return self._storage.createMatchingCache(function_ids)
-
-    ######## Below this line, nothing will be overwritten by Subclasses #########
-
-    def _getMatchesRoutine(self):
-        # Query, VS, Sample
-        # All use this version
-        pichash_matches = self._harmonizePicHashMatches(self._getPicHashMatches())
-        candidate_groups = self._createMinHashCandidateGroups(pichash_matches)
-        minhash_matches = self._harmonizeMinHashMatches(self._sample_id, self._performMinHashMatching(candidate_groups))
-        matching_report = self._craftResultDict(pichash_matches, minhash_matches)
-        return matching_report
-
-    # Reports PROGRESS
-    def _performMinHashMatching(self, candidate_groups: Dict[int, Set[int]]) -> List[Tuple[int, int, int, int, float]]:
-        # Query, VS, Sample
-        # All use the same version
-        """perform matching between candidates, provided as candidate_groups in the form of a dict: {function_id: [function_ids]}"""
-        matching_results: List[Tuple[int, int, int, int, float]] = []
         cache_function_ids = set()
         for candidate_key, candidate_functions in candidate_groups.items():
             cache_function_ids.add(candidate_key)
             cache_function_ids.update(candidate_functions)
         LOGGER.info("creating cache for %d functions", len(cache_function_ids))
-        cache = self._createMatchingCache(cache_function_ids)
+        return self._storage.createMatchingCache(cache_function_ids)
+
+    ######## Below this line, nothing will be overwritten by Subclasses #########
+
+    def filter_pichashes_from_candidate_groups(self, matching_cache, candidate_groups, pichash_matches):
+        finished_tuples = set()
+        for pichash_match in pichash_matches:
+            own_fid, foreign_sid, foreign_fid = pichash_match
+            if own_fid in candidate_groups:
+                # remove all candidates of the respective sample instead
+                if (own_fid, foreign_sid) not in finished_tuples:
+                    if matching_cache.isSampleId(foreign_sid):
+                        pichash_match_sid_fids = matching_cache.getFunctionIdsBySampleId(foreign_sid)
+                        candidate_groups[own_fid].difference_update(pichash_match_sid_fids)
+                        finished_tuples.add((own_fid, foreign_sid))
+        return candidate_groups
+
+    def _getMatchesRoutine(self):
+        # Query, VS, Sample
+        # All use this version
+        pichash_matches = self._harmonizePicHashMatches(self._getPicHashMatches())
+        LOGGER.info("Calculated PicHash matches")
+        all_minhash_matches = {}
+        # if we have an exceedingly large number of functions, we need to process in batches...
+        for start_index in range(0, len(self._function_entries), 10000):
+            candidate_groups = self._createMinHashCandidateGroups(start=start_index, end=start_index+10000)
+            LOGGER.info("Created candidate groups from MinHash bands")
+            matching_cache = self._createMatchingCache(candidate_groups)
+            LOGGER.info("Created MatchingCache")
+            if self._worker._minhash_config.PICHASH_IMPLIES_MINHASH_MATCH:
+                LOGGER.info("Removing PicHash matches")
+                candidate_groups = self.filter_pichashes_from_candidate_groups(matching_cache, candidate_groups, pichash_matches)
+                LOGGER.info("Removed PicHash matches from CandidateGroups")
+            LOGGER.info("Now starting MinHash matching")
+            minhash_matches = self._harmonizeMinHashMatches(self._sample_id, self._performMinHashMatching(candidate_groups, matching_cache))
+            all_minhash_matches.update(minhash_matches)
+        LOGGER.info("Calculated MinHash matches.")
+        matching_report = self._craftResultDict(pichash_matches, all_minhash_matches)
+        LOGGER.info("Returning aggregated match report.")
+        return matching_report
+
+    # Reports PROGRESS
+    def _performMinHashMatching(self, candidate_groups: Dict[int, Set[int]], cache) -> List[Tuple[int, int, int, int, float]]:
+        # Query, VS, Sample
+        # All use the same version
+        """perform matching between candidates, provided as candidate_groups in the form of a dict: {function_id: [function_ids]}"""
+        # result format: sample_id_a, function_id_a, sample_id_b, function_id_b, score
+        matching_results: List[Tuple[int, int, int, int, float]] = []
+        organized_matching_results: Dict[Tuple[int, int, int], Tuple[int, float]] = defaultdict(lambda: (-1, -1.0))
         packed_tuples = self._unrollGroupsAsPackedTuples(cache, candidate_groups)
         num_packed_tuples = self._countPackedTuples(candidate_groups)
         self._progress_reporter.set_total(num_packed_tuples)
+        calculation_function = functools.partial(
+            self._worker.minhasher.calculateScoresFromPackedTuples, ignore_threshold=True, minhash_threshold=self._minhash_threshold
+        )
+        packed_tuples = [p for p in packed_tuples]
+        counted_scores = Counter()
         if self._worker._minhash_config.MINHASH_POOL_MATCHING:
-            target_function = functools.partial(
-                self._worker.minhasher.calculateScoresFromPackedTuples, minhash_threshold=self._minhash_threshold
-            )
             with Pool(cpu_count()) as pool:
                 for pool_result in tqdm.tqdm(
-                    pool.imap_unordered(target_function, packed_tuples),
+                    pool.imap_unordered(calculation_function, packed_tuples),
                     total=num_packed_tuples,
                 ):
-                    matching_results.extend(pool_result)
+                    # for key, new_value in pool_result.items():
+                    for single_result in pool_result:
+                        sample_id_a, function_id_a, sample_id_b, function_id_b, score = single_result
+                        counted_scores[score] += 1
+                        if score > self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_THRESHOLD:
+                            key = (sample_id_a, function_id_a, sample_id_b)
+                            new_value = (function_id_b, score)
+                            original_value = organized_matching_results[key]
+                            organized_matching_results[key] = max([original_value, new_value], key=lambda x:x[1])
                     self._progress_reporter.step()
         else:
             packed_tuple: List[Tuple[int, int, bytes, int, int, bytes]]
             for packed_tuple in tqdm.tqdm(packed_tuples, total=num_packed_tuples):
-                matching_results.extend(
-                    self._worker.minhasher.calculateScoresFromPackedTuples(
-                        packed_tuple, minhash_threshold=self._minhash_threshold
-                    )
-                )
+                pool_result = calculation_function(packed_tuple)
+                for key, new_value in pool_result.items():
+                    original_value = organized_matching_results[key]
+                    organized_matching_results[key] = max([original_value, new_value], key=lambda x:x[1])
                 self._progress_reporter.step()
+        full_score_counts = sorted([(item[0]*64/100, item[1]) for item in dict(counted_scores).items()])
+        LOGGER.info("Minhash Signature Field Match Counts: " + ", ".join([f"({i[0]}: {i[1]})" for i in full_score_counts]))
+        matching_results = [k+v for k, v in organized_matching_results.items()]
         self._storage.clearMatchingCache()
         return matching_results
 
-    def _countPackedTuples(self, candidate_pairs, packsize=10000) -> int:
+    def _countPackedTuples(self, candidate_pairs, packsize=20000) -> int:
         count = 0
         for candidate_ids in candidate_pairs.values():
             count += len(candidate_ids) 
         quotient, remainder = divmod(count, packsize)
+        LOGGER.info("Processing a total of %d candidates.", count)
         return quotient + int(bool(remainder)) # always round up
 
     def _unrollGroupsAsPackedTuples(
-        self, cache: Union["MatchingCache", "MemoryStorage"], candidate_pairs, packsize=10000
+        self, cache: Union["MatchingCache", "MemoryStorage"], candidate_pairs, packsize=20000
     ) -> Iterable[List[Tuple[int, int, bytes, int, int, bytes]]]:
         # Query, VS, Sample
         # All were identical
@@ -309,17 +350,17 @@ class MatcherInterface(object):
                 self_matched_functions.add(own_function_id)
                 self_matched_functions.add(foreign_function_id)
             else:
-                if not foreign_sample_id in self._sample_to_lib_info:
+                if foreign_sample_id not in self._sample_to_lib_info:
                     self._sample_to_lib_info[foreign_sample_id] = (
                         self._storage.getLibraryInfoForSampleId(foreign_sample_id) is not None
                     )
                 has_libinfo = self._sample_to_lib_info[foreign_sample_id]
 
-                if not foreign_sample_id in self._sample_id_to_entry:
+                if foreign_sample_id not in self._sample_id_to_entry:
                     self._sample_id_to_entry[foreign_sample_id] = self._storage.getSampleById(foreign_sample_id)
                 foreign_family_id = self._sample_id_to_entry[foreign_sample_id].family_id
 
-                if not own_function_id in match_function_mapping:
+                if own_function_id not in match_function_mapping:
                     match_function_mapping[own_function_id] = {
                         "num_bytes": sample_fid_to_binweight[own_function_id],
                         "offset": sample_fid_to_offset[own_function_id],
@@ -401,6 +442,7 @@ class MatcherInterface(object):
                 "filename": sample_info.filename,
                 "sample_id": foreign_sample_id,
                 "num_bytes": sample_info.binweight,
+                "is_library": sample_info.is_library,
                 "num_functions": sample_info.statistics["num_functions"],
                 "matched": {
                     "functions": {
@@ -494,6 +536,7 @@ class MatcherInterface(object):
         for key, new_entry in pichash_matches.items():
             if key in all_matches:
                 old_entry = all_matches[key]
+                # update values: score, is_pic, is_min
                 all_matches[key] = tuple([max(old_entry[i], new_entry[i]) for i in range(3)])
             else:
                 all_matches[key] = new_entry
