@@ -21,8 +21,12 @@ def getMatchingParams(req_params):
             if key == "force_recalculation":
                 if value.lower() == "true":
                     parameters["force_recalculation"] = True
+            if key == "band_matches_required":
+                band_matches_required = int(value)
+                band_matches_required = max(0, band_matches_required)
+                parameters["band_matches_required"] = band_matches_required
         except:
-            LOGGER.warn(f"Failed to handle request parameter: {key}: {value}")
+            LOGGER.warning(f"Failed to handle request parameter: {key}: {value}")
     return parameters
 
 
