@@ -13,13 +13,15 @@ from mcrit.index.SearchCursor import FullSearchCursor
 from mcrit.index.SearchQueryTree import AndNode, BaseVisitor, FilterSingleElementLists, NodeType, OrNode, PropagateNot, SearchConditionNode, SearchFieldResolver
 from mcrit.storage.FamilyEntry import FamilyEntry
 from mcrit.storage.FunctionEntry import FunctionEntry
+from mcrit.storage.MatchingCache import MatchingCache
 from mcrit.storage.SampleEntry import SampleEntry
 from mcrit.storage.StorageInterface import StorageInterface
 
 if TYPE_CHECKING: # pragma: no cover
+    from mcrit.config.McritConfig import McritConfig
     from mcrit.config.StorageConfig import StorageConfig
+    from mcrit.config.MinHashConfig import MinHashConfig
     from mcrit.minhash.MinHash import MinHash
-    from mcrit.storage.MatchingCache import MatchingCache
     from smda.common.SmdaFunction import SmdaFunction
     from smda.common.SmdaReport import SmdaReport
 
@@ -111,6 +113,8 @@ class MemoryStorage(StorageInterface):
     _families: Dict[int, FamilyEntry]
     _samples: Dict[int, "SampleEntry"]
     _functions: Dict[int, "FunctionEntry"]
+    _query_samples: Dict[int, "SampleEntry"]
+    _query_functions: Dict[int, "FunctionEntry"]
     _pichashes: Dict[int, Set[Tuple[int, int, int]]]
     # Dict[band_number, Dict[?, List[function_id]]]
     _bands: Dict[int, Dict[int, List[int]]]
@@ -118,7 +122,7 @@ class MemoryStorage(StorageInterface):
     _sample_by_sha256: Dict[str, int]
     _sample_id_to_function_ids: Dict[int, List[int]]
 
-    def __init__(self, config: "StorageConfig") -> None:
+    def __init__(self, config: "McritConfig") -> None:
         super().__init__(config)  # sets config
         self._setupEmptyStorage()
         self.blockhasher = BlockHasher()
@@ -132,9 +136,16 @@ class MemoryStorage(StorageInterface):
         self._families = {}
         self._samples = {}
         self._functions = {}
+        self._query_samples = {}
+        self._query_functions = {}
         self._pichashes = {}
-        self._bands = {band_number: {} for band_number in range(self._config.STORAGE_NUM_BANDS)}
+        self._bands = {band_number: {} for band_number in range(self._storage_config.STORAGE_NUM_BANDS)}
         self._counters = defaultdict(lambda: 0)
+        # initialize query sample/function ids 
+        if self._counters["query_samples"] == 0:
+            self._counters["query_samples"] += 1
+        if self._counters["query_functions"] == 0:
+            self._counters["query_functions"] += 1
         # caches
         self._sample_by_sha256 = {}
         self._sample_id_to_function_ids = defaultdict(list)
@@ -149,7 +160,6 @@ class MemoryStorage(StorageInterface):
         self._counters[name] += 1
         return result
 
-
     def _updateFamilyStats(self, family_id, num_samples_inc, num_functions_inc, num_library_samples_inc):
         family_entry = self.getFamily(family_id)
         assert family_entry is not None
@@ -162,16 +172,22 @@ class MemoryStorage(StorageInterface):
         if not self.isSampleId(sample_id):
             return False
         function_ids = self._sample_id_to_function_ids[sample_id]
+        if sample_id < 0:
+            for function_id in function_ids:
+                del self._functions[function_id]
+            del self._sample_id_to_function_ids[sample_id]
+            del self._samples[sample_id]
+            return True
         for function_id in function_ids:
             function_entry = self._functions[function_id]
+            minhash = function_entry.getMinHash(self._minhash_config.MINHASH_SIGNATURE_BITS)
             # remove function
             del self._functions[function_id]
             # remove pichash entries
             if function_entry.pichash:
                 self._pichashes[function_entry.pichash].remove((function_entry.family_id, sample_id, function_id))
-            minhash = function_entry.getMinHash()
             # remove minhash entries, if necessary
-            if not minhash:
+            if not minhash or not minhash.hasMinHash():
                 continue
             band_hashes = self.getBandHashesForMinHash(minhash)
             for band_number, band_hash in sorted(band_hashes.items()):
@@ -193,12 +209,93 @@ class MemoryStorage(StorageInterface):
         del self._samples[sample_id]
         return True
 
+    def modifySample(self, sample_id: int, update_information: dict) -> bool:
+        if not self.isSampleId(sample_id):
+            return False
+        if "is_library" in update_information:
+            is_library_info_changed = self._samples[sample_id].is_library != update_information["is_library"]
+            self._samples[sample_id].is_library = update_information["is_library"]
+            if is_library_info_changed:
+                self._families[family_id].num_library_samples += 1 if update_information["is_library"] else -1
+        if "family_name" in update_information:
+            family_name = update_information["family_name"]
+            old_family_id = self._samples[sample_id].family_id
+            family_id = self.addFamily(family_name)
+            # update sample_entry and function_entries with new family information
+            self._samples[sample_id].family = family_name
+            self._samples[sample_id].family_id = family_id
+            for function_id, function_entry in self._functions.items():
+                if function_entry.sample_id == sample_id:
+                    function_entry.family_id = family_id
+                    self._functions[function_id] = function_entry  
+                    self._pichashes[function_entry.pichash].remove((old_family_id, sample_id, function_id))
+                    self._pichashes[function_entry.pichash].add((family_id, sample_id, function_id))
+            # update family information itself
+            self._families[old_family_id].num_samples -= 1
+            self._families[old_family_id].num_functions -= self._samples[sample_id].statistics["num_functions"]
+            self._families[family_id].num_samples += 1
+            self._families[family_id].num_functions += self._samples[sample_id].statistics["num_functions"]
+            if self._samples[sample_id].is_library:
+                self._families[old_family_id].num_library_samples -= 1
+                self._families[family_id].num_library_samples += 1
+            if self._families[old_family_id].num_samples == 0 and old_family_id != 0:
+                self._families.pop(old_family_id)
+        if "version" in update_information:
+            self._samples[sample_id].version = update_information["version"]
+        if "component" in update_information:
+            self._samples[sample_id].component = update_information["component"]
+        return True
+
+    def modifyFamily(self, family_id: int, update_information: dict) -> bool:
+        if not self.isFamilyId(family_id):
+            return False
+        old_family_info = self.getFamily(family_id)
+        if "is_library" in update_information:
+            for sample_id, sample_entry in self._samples.items():
+                if family_id == sample_entry.family_id:
+                    self._samples["sample_id"].is_library = update_information["is_library"]
+            self._families[family_id].num_library_samples = self._families[family_id].num_samples
+        if "family_name" in update_information:
+            old_family_info = self.getFamily(family_id)
+            family_name = update_information["family_name"]
+            new_family_id = self.addFamily(family_name)
+            new_family_info = self.getFamily(new_family_id)
+            new_num_samples = new_family_info.num_samples + old_family_info.num_samples
+            new_num_functions = new_family_info.num_functions + old_family_info.num_functions
+            new_num_lib_samples = new_family_info.num_library_samples + old_family_info.num_library_samples
+            # update family_entry
+            if family_id == 0:
+                self._families[0].num_samples = 0
+                self._families[0].num_functions = 0
+                self._families[0].num_library_samples = 0
+            else:
+                self._families.pop(family_id)
+            self._families[new_family_id].num_samples = new_num_samples
+            self._families[new_family_id].num_functions = new_num_functions
+            self._families[new_family_id].num_library_samples = new_num_lib_samples
+            # update sample_entry and function_entries with new family information
+            for sample_id, sample_entry in self._samples.items():
+                if family_id == sample_entry.family_id:
+                    self._samples[sample_id].family_id = new_family_id
+                    self._samples[sample_id].family = family_name
+            for function_id, function_entry in self._functions.items():
+                if family_id == function_entry.family_id:
+                    self._functions[function_id].family_id = new_family_id
+                    self._pichashes[function_entry.pichash].remove((family_id, sample_id, function_id))
+                    self._pichashes[function_entry.pichash].add((new_family_id, sample_id, function_id))
+        self._updateDbState()
+        return True
 
     def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
-        if family_id not in self.families:
+        if family_id not in self._families:
             return False
         sample_entries = self.getSamplesByFamilyId(family_id)
-        self._families.pop(family_id)
+        if family_id == 0:
+            self._families[0].num_samples = 0
+            self._families[0].num_functions = 0
+            self._families[0].num_library_samples = 0
+        else:
+            self._families.pop(family_id)
         if keep_samples:
             for sample_entry in sample_entries:
                 self._samples[sample_entry.sample_id]["family_id"] = 0
@@ -209,29 +306,43 @@ class MemoryStorage(StorageInterface):
                     function_ids_to_modify.add(function_id)
             for function_id in function_ids_to_modify:
                     self._functions[function_id]["family_id"] = 0
+            self._families[0].num_samples += len(sample_entries)
+            self._families[0].num_functions += len(function_ids_to_modify)
+            self._families[0].num_library_samples += len([s for s in sample_entries if s.is_library])
         else:
             for sample_entry in sample_entries:
                 self.deleteSample(sample_entry.sample_id)
         self._updateDbState()
         return True
 
-    def addSmdaReport(self, smda_report: "SmdaReport") -> Optional["SampleEntry"]:
+    def addSmdaReport(self, smda_report: "SmdaReport", isQuery=False) -> Optional["SampleEntry"]:
         sample_entry = None
-        if not self.getSampleBySha256(smda_report.sha256):
-            family_id = self.addFamily(smda_report.family)
+        if isQuery:
             sample_entry = SampleEntry(
-                smda_report, sample_id=self._useCounter("samples"), family_id=family_id
+                smda_report, sample_id=-1 * self._useCounter("query_samples"), family_id=0
             )
-            self._samples[sample_entry.sample_id] = sample_entry
-            self._sample_by_sha256[sample_entry.sha256] = sample_entry.sample_id
+            self._query_samples[sample_entry.sample_id] = sample_entry
             function_ids = []
             for smda_function in smda_report.getFunctions():
-                function_entry = self._addFunction(sample_entry, smda_function)
+                function_entry = self._addFunction(sample_entry, smda_function, isQuery=True)
                 function_ids.append(function_entry.function_id)
             self._sample_id_to_function_ids[sample_entry.sample_id] = function_ids
-            self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
         else:
-            LOGGER.warn("Sample %s already existed, skipping.", smda_report.sha256)
+            if not self.getSampleBySha256(smda_report.sha256):
+                family_id = self.addFamily(smda_report.family)
+                sample_entry = SampleEntry(
+                    smda_report, sample_id=self._useCounter("samples"), family_id=family_id
+                )
+                self._samples[sample_entry.sample_id] = sample_entry
+                self._sample_by_sha256[sample_entry.sha256] = sample_entry.sample_id
+                function_ids = []
+                for smda_function in smda_report.getFunctions():
+                    function_entry = self._addFunction(sample_entry, smda_function)
+                    function_ids.append(function_entry.function_id)
+                self._sample_id_to_function_ids[sample_entry.sample_id] = function_ids
+                self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
+            else:
+                LOGGER.warning("Sample %s already existed, skipping.", smda_report.sha256)
         return sample_entry
 
     def importSampleEntry(self, sample_entry: "SampleEntry") -> Optional["SampleEntry"]:
@@ -242,7 +353,7 @@ class MemoryStorage(StorageInterface):
             self._sample_by_sha256[sample_entry.sha256] = sample_id
             self._updateFamilyStats(sample_entry.family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
         else:
-            LOGGER.warn("Sample %s already existed, skipping.", sample_entry.sha256)
+            LOGGER.warning("Sample %s already existed, skipping.", sample_entry.sha256)
         return sample_entry
 
     def importFunctionEntry(self, function_entry: "FunctionEntry") -> Optional["FunctionEntry"]:
@@ -281,7 +392,7 @@ class MemoryStorage(StorageInterface):
         return {"family": sample_entry.family, "version": sample_entry.version} if sample_entry.is_library else None
 
     def _addFunction(
-        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None
+        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None, isQuery=False
     ) -> "FunctionEntry":
         """Add a function (and optionally its MinHash) to storage, using the respective SampleEntry for reference.
 
@@ -293,7 +404,12 @@ class MemoryStorage(StorageInterface):
         Returns:
             A FunctionEntry
         """
-        function_entry = FunctionEntry(sample_entry, smda_function, self._useCounter("functions"), minhash=minhash)
+        if isQuery:
+            function_entry = FunctionEntry(sample_entry, smda_function, -1 * self._useCounter("query_functions"), minhash=minhash)
+            self._query_functions[function_entry.function_id] = function_entry
+            return function_entry
+        else:
+            function_entry = FunctionEntry(sample_entry, smda_function, self._useCounter("functions"), minhash=minhash)
         image_lower = sample_entry.base_addr
         image_upper = image_lower + sample_entry.binary_size
         picblockhashes = []
@@ -324,9 +440,29 @@ class MemoryStorage(StorageInterface):
         for minhash in minhashes:
             self.addMinHash(minhash)
 
-    def createMatchingCache(self, function_ids: List[int]) -> "MemoryStorage":
-        """We are already memory-only and won't gain anything using a cache"""
-        return self
+    def createMatchingCache(self, function_ids: List[int]) -> MatchingCache:
+        # TODO: we might want add a flag to allow/disallow returning self
+        cache_data = self._getCacheDataForFunctionIds(function_ids)
+        return MatchingCache(cache_data)
+
+    def _getCacheDataForFunctionIds(self, function_ids: List[int]) -> Dict:
+        cache_data = {}
+        sample_ids = {}
+        sample_to_func_ids = {}
+        minhashes = {}
+        for function_id in function_ids:
+            function_entry = self._functions[function_id]
+            function_id = function_entry.function_id
+            sample_id = function_entry.sample_id
+            minhashes[function_id] = function_entry.minhash
+            sample_ids[function_id] = sample_id
+            if sample_id not in sample_to_func_ids:
+                sample_to_func_ids[sample_id] = set()
+            sample_to_func_ids[sample_id].add(function_id)
+        cache_data["func_id_to_minhash"] = minhashes
+        cache_data["func_id_to_sample_id"] = sample_ids
+        cache_data["sample_id_to_func_ids"] = sample_to_func_ids
+        return cache_data
 
     def clearMatchingCache(self) -> None:
         """no dedicated cache - no cleanup"""
@@ -348,10 +484,10 @@ class MemoryStorage(StorageInterface):
         return deepcopy(list(self._families.keys()))
 
     def isSampleId(self, sample_id: int) -> bool:
-        return sample_id in self._samples
+        return sample_id in self._samples or sample_id in self._query_samples
 
     def isFunctionId(self, function_id: int) -> bool:
-        return function_id in self._functions
+        return function_id in self._functions or function_id in self._query_functions
 
     def isFamilyId(self, family_id: int) -> bool:
         return family_id in self._families
@@ -359,13 +495,12 @@ class MemoryStorage(StorageInterface):
     def isPicHash(self, pichash: int) -> bool:
         return pichash in self._pichashes
 
-    def getMatchesForPicHash(self, pichash: int) -> Optional[Set[Tuple[int, int, int]]]:
+    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int, int]]:
         if pichash not in self._pichashes:
             return set()
         return deepcopy(self._pichashes[pichash])
 
-
-    def getMatchesForPicBlockHash(self, picblockhash: int) -> Optional[Set[Tuple[int, int, int, int]]]:
+    def getMatchesForPicBlockHash(self, picblockhash: int) -> Set[Tuple[int, int, int, int]]:
         result = set()
         for function_id, function_entry in self._functions.items():
             for pbh in function_entry.picblockhashes:
@@ -392,8 +527,14 @@ class MemoryStorage(StorageInterface):
         return family_id
 
     def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
-        if function_id in self._functions:
-            function_entry = deepcopy(self._functions[function_id])
+        function_entry = None
+        if function_id < 0:
+            if function_id in self._query_functions:
+                function_entry = deepcopy(self._query_functions[function_id])
+        else:
+            if function_id in self._functions:
+                function_entry = deepcopy(self._functions[function_id])
+        if function_entry:
             if with_xcfg is False:
                 function_entry.xcfg = None
             return deepcopy(function_entry)
@@ -401,16 +542,25 @@ class MemoryStorage(StorageInterface):
 
     # TODO does this need to be more efficient?
     def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
-        if sample_id in self._samples:
+        if sample_id in self._samples or sample_id in self._query_samples:
             function_ids = self._sample_id_to_function_ids[sample_id]
             function_entries = []
             for function_id in function_ids:
-                function_entry = self._functions[function_id]
+                if sample_id < 0:
+                    function_entry = self._query_functions[function_id]
+                else:
+                    function_entry = self._functions[function_id]
                 function_entries.append(function_entry)
             # TODO is deepcopy necessary?
             return deepcopy(function_entries)
         return None
 
+    def getFunctionIdsBySampleId(self, sample_id: int) -> Optional[List["int"]]:
+        function_ids = None
+        if sample_id in self._samples or sample_id in self._query_samples:
+            function_ids = self._sample_id_to_function_ids[sample_id]
+        return function_ids
+
     def getFunctions(self, start_index: int, limit: int) -> Optional["FunctionEntry"]:
         index = 0
         function_entries = []
@@ -426,6 +576,8 @@ class MemoryStorage(StorageInterface):
     def getSampleById(self, sample_id: int) -> Optional["SampleEntry"]:
         if sample_id in self._samples:
             return deepcopy(self._samples[sample_id])
+        if sample_id in self._query_samples:
+            return deepcopy(self._query_samples[sample_id])
 
     def getSampleBySha256(self, sha256: str) -> Optional["SampleEntry"]:
         sample_entry = None
@@ -438,6 +590,8 @@ class MemoryStorage(StorageInterface):
         sample_id = None
         if function_id in self._functions:
             sample_id = self._functions[function_id].sample_id
+        if function_id in self._query_functions:
+            sample_id = self._query_functions[function_id].sample_id
         return sample_id
 
     def getSamples(self, start_index: int, limit: int) -> Optional["SampleEntry"]:
@@ -485,33 +639,42 @@ class MemoryStorage(StorageInterface):
 
     # -> Dict[function_id, Set[function_id]]
     # TODO optimize or move to interface
-    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"]) -> Dict[int, Set[int]]:
+    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"], band_matches_required=1) -> Dict[int, Set[int]]:
         candidates = {}
         for function_id, minhash in function_id_to_minhash.items():
-            candidates[function_id] = self.getCandidatesForMinHash(minhash)
+            candidates[function_id] = self.getCandidatesForMinHash(minhash, band_matches_required=band_matches_required)
         return candidates
 
     # -> Set[function_id]
-    def getCandidatesForMinHash(self, minhash: "MinHash") -> Set[int]:
-        candidates = set([])
+    def getCandidatesForMinHash(self, minhash: "MinHash", band_matches_required=1) -> Set[int]:
+        candidates = {}
         band_hashes = self.getBandHashesForMinHash(minhash)
         for band_number, band_hash in sorted(band_hashes.items()):
             if band_hash in self._bands[band_number]:
-                candidates.update(self._bands[band_number][band_hash])
-        return candidates
+                for function_id in self._bands[band_number][band_hash]:
+                    if function_id not in candidates:
+                        candidates[function_id] = 0
+                    candidates[function_id] += 1
+        # reduce candidates based on banding requirements
+        valid_candidates = set([])
+        for function_id, hit_count in candidates.items():
+            if hit_count >= band_matches_required:
+                valid_candidates.add(function_id)
+        return valid_candidates
 
-    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None) -> List["FunctionEntry"]:
+    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None, only_function_ids=False) -> List["FunctionEntry"]:
+        result = []
         if function_ids is None:
-            return [
-                function_entry
-                for _, function_entry in self._functions.items()
-                if function_entry.xcfg and not function_entry.minhash
-            ]
-        return [
-            function_entry
-            for _, function_entry in self._functions.items()
-            if (not function_entry.minhash and function_entry.function_id in function_ids)
-        ]
+            if only_function_ids:
+                result = [function_entry.function_id for _, function_entry in self._functions.items() if function_entry.xcfg and not function_entry.minhash]
+            else:
+                result = [function_entry for _, function_entry in self._functions.items() if function_entry.xcfg and not function_entry.minhash]
+            return result
+        if only_function_ids:
+            result = [function_entry.function_id for _, function_entry in self._functions.items() if (not function_entry.minhash and function_entry.function_id in function_ids)]
+        else:
+            result = [function_entry for _, function_entry in self._functions.items() if (not function_entry.minhash and function_entry.function_id in function_ids)]
+        return result
 
     def deleteXcfgData(self) -> None:
         for function_id, function_entry in self._functions.items():
@@ -569,8 +732,20 @@ class MemoryStorage(StorageInterface):
             self._bands[band_number][band_hash].append(minhash.function_id)
 
 
-    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None) -> Dict:
+    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None, progress_reporter=None) -> Dict:
         # query once to get all blocks from the functions of our samples
+        block_statistics = {
+            "by_sample_id": {
+                sample_id: {
+                    "sample_id": sample_id,
+                    "total_blocks": 0,
+                    "characteristic_blocks": 0,
+                    "unique_blocks": 0
+                } for sample_id in sample_ids
+            },
+            "unique_blocks_overall": 0,
+            "num_samples": len(sample_ids)
+        }
         candidate_picblockhashes = {}
         for function_id, entry in self._functions.items():
             sample_id = entry.sample_id
@@ -585,12 +760,24 @@ class MemoryStorage(StorageInterface):
                         "instructions": []
                     }
                 candidate_picblockhashes[block_hash]["samples"].add(sample_id)
+        # update statistics based on candidates
+        for picblockhash, entry in candidate_picblockhashes.items():
+            for sample_id in entry["samples"]:
+                block_statistics["by_sample_id"][sample_id]["total_blocks"] += 1
         LOGGER.info(f"Found {len(candidate_picblockhashes)} candidate picblock hashes")
         for functiond_id, entry in self._functions.items():
             sample_id = entry.sample_id
             if sample_id not in sample_ids:
                 for block_entry in entry.picblockhashes:
                     candidate_picblockhashes.pop(block_entry["hash"], None)
+        # update statistics again after having reduced to results
+        for picblockhash, entry in candidate_picblockhashes.items():
+            if len(entry["samples"]) == 1:
+                single_sample_id = list(entry["samples"])[0]
+                block_statistics["by_sample_id"][single_sample_id]["unique_blocks"] += 1
+            for sample_id in entry["samples"]:
+                block_statistics["by_sample_id"][sample_id]["characteristic_blocks"] += 1
+        block_statistics["unique_blocks_overall"] = len(candidate_picblockhashes)
         LOGGER.info(f"Reduced to {len(candidate_picblockhashes)} unique picblock hashes")
         # iterate over candidates by function_id and extract instructions
         function_id_to_block_offsets = {}
@@ -604,7 +791,7 @@ class MemoryStorage(StorageInterface):
                 continue
             for block_offset, picblockhash in function_id_to_block_offsets[function_id]:
                 candidate_picblockhashes[picblockhash]["instructions"] = entry.xcfg["blocks"][str(block_offset)]
-        return candidate_picblockhashes
+        return {"statistics": block_statistics, "unique_blocks": candidate_picblockhashes}
 
     ##### helpers for search ######
 
