@@ -4,6 +4,7 @@ import json
 import logging
 import datetime
 import traceback
+from collections import defaultdict
 from operator import itemgetter
 from itertools import zip_longest
 from typing import Any, TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Tuple, Union
@@ -13,7 +14,7 @@ try:
     from pymongo import InsertOne, MongoClient, UpdateOne
     from pymongo.collection import ReturnDocument
 except:
-    LOGGER.warn("pymongo package import failed - MongoDB backend will not be available.")
+    LOGGER.warning("pymongo package import failed - MongoDB backend will not be available.")
 
 from picblocks.blockhasher import BlockHasher
 
@@ -107,7 +108,7 @@ class MongoDbStorage(StorageInterface):
         super().__init__(config)  # sets config
         self._matching_cache = None
         self.blockhasher = BlockHasher()
-        self._initDb(self._config.STORAGE_SERVER, self._config.STORAGE_PORT, self._config.STORAGE_MONGODB_DBNAME)
+        self._initDb(self._storage_config.STORAGE_SERVER, self._storage_config.STORAGE_PORT, self._storage_config.STORAGE_MONGODB_DBNAME)
 
     def _initDb(self, server, port, db_name):
         self._database = MongoClient(server, port=port)[db_name]
@@ -123,10 +124,25 @@ class MongoDbStorage(StorageInterface):
         self._database["functions"].create_index("_pichash")
         self._database["functions"].create_index("_picblockhashes.hash")
         self._database["functions"].create_index("_picblockhashes.offset")
+        # stored without guarantee of existence
+        self._database["query_samples"].create_index("sample_id")
+        self._database["query_functions"].create_index("function_id")
+        self._database["query_functions"].create_index("sample_id")
+        # ensure that their counters are at least 1, so that they never contain items with sample_id/function_id 0
+        self._database.counters.find_one_and_update(
+            filter={"name": "query_samples", "value": 0}, 
+            update={"$inc": {"value": 1}}, 
+            upsert=True
+        )
+        self._database.counters.find_one_and_update(
+            filter={"name": "query_functions", "value": 0}, 
+            update={"$inc": {"value": 1}}, 
+            upsert=True
+        )
         self._database["matches"].create_index("match_id")
         self._database["candidates"].create_index("function_id")
         self._database["counters"].create_index("name")
-        for band_id in range(self._config.STORAGE_NUM_BANDS):
+        for band_id in range(self._storage_config.STORAGE_NUM_BANDS):
             self._database["band_%d" % band_id].create_index("band_hash")
         # Add Family "" if it is not already in storage
         if self.getFamily(0) is None:
@@ -174,12 +190,14 @@ class MongoDbStorage(StorageInterface):
         except Exception as exc:
             self._dbLogError(
                 'Database insert for collection "%s" failed.' % collection,
-                {"user": "internal/mcrit", "traceback": traceback.format_exc(exc).split("\n")},
+                {"user": "internal/mcrit", "traceback": traceback.format_exc().split("\n")},
             )
             raise ValueError("Database insert failed.")
 
 
     def _dbInsertMany(self, collection: str, data: List["Dict"]):
+        if len(data) == 0:
+            return []
         try:
             insert_result = self._database[collection].insert_many([self._toBinary(document) for document in data])
             if insert_result.acknowledged:
@@ -188,7 +206,7 @@ class MongoDbStorage(StorageInterface):
         except Exception as exc:
             self._dbLogError(
                 'Database insert_many for collection "%s" failed.' % collection,
-                {"user": "internal/mcrit", "traceback": traceback.format_exc(exc).split("\n")},
+                {"user": "internal/mcrit", "traceback": traceback.format_exc().split("\n")},
             )
             raise ValueError("Database insert failed.")
 
@@ -201,7 +219,7 @@ class MongoDbStorage(StorageInterface):
         except Exception as exc:
             self._dbLogError(
                 'Database query for collection "%s" failed.' % collection,
-                {"user": "internal/mcrit", "traceback": traceback.format_exc(exc).split("\n")},
+                {"user": "internal/mcrit", "traceback": traceback.format_exc().split("\n")},
             )
             raise ValueError("Database query failed.")
 
@@ -322,42 +340,55 @@ class MongoDbStorage(StorageInterface):
 
     # NOTE this actually only gets used in MemoryStorage / MatchingCache
     def getSampleIdByFunctionId(self, function_id: int) -> Optional[int]:
-        function_document = self._database.functions.find_one({"function_id": function_id})
+        if function_id < 0:
+            function_document = self._database.query_functions.find_one({"function_id": function_id})
+        else:
+            function_document = self._database.functions.find_one({"function_id": function_id})
         if function_document is None:
             return None
         return function_document["sample_id"]
 
     def deleteSample(self, sample_id: int) -> bool:
-        # TODO the way of removing function_ids from band hashes here is super inefficient and slow
+        sample_entry = self.getSampleById(sample_id)
+        if sample_entry is None:
+            return False
+        if sample_id < 0:
+            # remove functions
+            self._database.query_functions.delete_many({"sample_id": sample_id})
+            # remove sample
+            self._database.query_samples.delete_one({"sample_id": sample_id})
+            return 
         function_entries = self.getFunctionsBySampleId(sample_id)
         if function_entries is None:
             # in this case sample_id is does not exist
             return False
 
+        # collect all band entries that need updating and pull all function_ids at once.
+        # might need to batch this into slices of function_ids again
+        minhashes_to_remove = {band_number: {} for band_number in range(self._storage_config.STORAGE_NUM_BANDS)}
         for function_entry in function_entries:
-            minhash = function_entry.getMinHash()
+            minhash = function_entry.getMinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
             # remove minhash entries, if necessary
             if not minhash or not minhash.hasMinHash():
                 continue
             band_hashes = self.getBandHashesForMinHash(minhash)
             for band_number, band_hash in sorted(band_hashes.items()):
-                # delete function id from bandhash
-                band_document = self._database["band_%d" % band_number].find_one_and_update(
-                    {"band_hash": band_hash},
-                    {"$pull": {"function_ids": minhash.function_id}},
-                    return_document=ReturnDocument.AFTER,
-                )
-                # delete bandhash if empty
-                if band_document is not None and len(band_document["function_ids"]) == 0:
-                    self._database["band_%d" % band_number].delete_one({"band_hash": band_hash})
+                if band_hash not in minhashes_to_remove[band_number]:
+                    minhashes_to_remove[band_number][band_hash] = []
+                if function_entry.function_id not in minhashes_to_remove[band_number][band_hash]:
+                    minhashes_to_remove[band_number][band_hash].append(function_entry.function_id)
+        self._updateBands(minhashes_to_remove, method="pull")
 
         # update family stats
-        sample_entry = self.getSampleById(sample_id)
         self._updateFamilyStats(sample_entry.family_id, -1, -sample_entry.statistics["num_functions"], -int(sample_entry.is_library))
         # remove functions
         self._database.functions.delete_many({"sample_id": sample_id})
         # remove sample
         self._database.samples.delete_one({"sample_id": sample_id})
+        # delete family if empty
+        family_info = self.getFamily(sample_entry.family_id)
+        if family_info.num_samples == 0 and family_info.family_id != 0:
+            self._database.families.delete_one({"family_id": family_info.family_id})
         self._updateDbState()
         return True
 
@@ -373,19 +404,101 @@ class MongoDbStorage(StorageInterface):
             }
         )
 
+    def modifySample(self, sample_id: int, update_information: dict) -> bool:
+        if not self.isSampleId(sample_id):
+            return False
+        sample_entry = self.getSampleById(sample_id)
+        if "is_library" in update_information:
+            is_library_info_changed = sample_entry.is_library != update_information["is_library"]
+            self._database.samples.update_one({"sample_id": sample_id}, {"$set": {"is_library": update_information["is_library"]}})
+            family_entry = self.getFamily(sample_entry.family_id)
+            if is_library_info_changed:
+                new_value = family_entry.num_library_samples + (1 if update_information["is_library"] else -1)
+                self._database.families.update_one({"family_id": sample_entry.family_id}, {"$set": {"num_library_samples": new_value}})
+        if "family_name" in update_information:
+            old_family_entry = self.getFamily(sample_entry.family_id)
+            family_name = update_information["family_name"]
+            family_id = self.addFamily(family_name)
+            new_family_entry = self.getFamily(family_id)
+            # update sample_entry and function_entries with new family information
+            self._database.samples.update_one({"sample_id": sample_id}, {"$set": {"family_id": family_id, "family": family_name}})
+            self._database.functions.update_many({"sample_id": sample_id}, {"$set": {"family_id": family_id}})
+            # update family entry with statistics
+            self._database.families.update_one(
+                {"family_id": old_family_entry.family_id}, 
+                {"$set": {
+                    "num_samples": old_family_entry.num_samples - 1,
+                    "num_functions": old_family_entry.num_functions - sample_entry.statistics["num_functions"],
+                    "num_library_samples": old_family_entry.num_library_samples - (1 if sample_entry.is_library else 0),
+                    }
+                }
+            )
+            self._database.families.update_one(
+                {"family_id": family_id}, 
+                {"$set": {
+                    "num_samples": new_family_entry.num_samples + 1,
+                    "num_functions": new_family_entry.num_functions + sample_entry.statistics["num_functions"],
+                    "num_library_samples": new_family_entry.num_library_samples + (1 if sample_entry.is_library else 0),
+                    }
+                }
+            )
+            old_family_entry = self.getFamily(sample_entry.family_id)
+            # delete family if empty
+            if old_family_entry.num_samples == 0 and old_family_entry.family_id != 0:
+                self._database.families.delete_one({"family_id": old_family_entry.family_id})
+            self._updateDbState()
+        if "version" in update_information:
+            self._database.samples.update_one({"sample_id": sample_id}, {"$set": {"version": update_information["version"]}})
+        if "component" in update_information:
+            self._database.samples.update_one({"sample_id": sample_id}, {"$set": {"component": update_information["component"]}})
+        return True
+
+    def modifyFamily(self, family_id: int, update_information: dict) -> bool:
+        if not self.isFamilyId(family_id):
+            return False
+        old_family_info = self.getFamily(family_id)
+        if "is_library" in update_information:
+            self._database.samples.update_many({"family_id": family_id}, {"$set": {"is_library": update_information["is_library"]}})
+            updated_count = old_family_info.num_samples if update_information["is_library"] else 0
+            self._database.families.update_one({"family_id": family_id}, {"$set": {"num_library_samples": updated_count}})
+            self._updateDbState()
+        if "family_name" in update_information:
+            old_family_info = self.getFamily(family_id)
+            family_name = update_information["family_name"]
+            new_family_id = self.addFamily(family_name)
+            new_family_info = self.getFamily(new_family_id)
+            new_num_samples = new_family_info.num_samples + old_family_info.num_samples
+            new_num_functions = new_family_info.num_functions + old_family_info.num_functions
+            new_num_lib_samples = new_family_info.num_library_samples + old_family_info.num_library_samples
+            # update family_entry
+            if family_id == 0:
+                self._database.families.update_one({"family_id": 0}, {"$set": {"num_samples": 0, "num_functions": 0, "num_library_samples": 0}})
+            else:
+                self._database.families.delete_one({"family_id": family_id})
+            self._database.families.update_one({"family_id": new_family_id}, {"$set": {"num_samples": new_num_samples, "num_functions": new_num_functions, "num_library_samples": new_num_lib_samples}})
+            # update sample_entry and function_entries with new family information
+            self._database.samples.update_many({"family_id": family_id}, {"$set": {"family_id": new_family_id, "family": family_name}})
+            self._database.functions.update_many({"family_id": family_id}, {"$set": {"family_id": new_family_id}})
+            self._updateDbState()
+        return True
+
     def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
-        family_document = self._database.families.find_one({"family_id": family_id})
-        family_name = family_document["family_name"]
-        if family_document is None:
+        family_entry = self.getFamily(family_id)
+        if family_entry is None:
             return False
         sample_entries = self.getSamplesByFamilyId(family_id)
-        self._database.families.delete_one({"family_id": family_id})
         if keep_samples:
             self._database.samples.update_many({"family_id": family_id}, {"$set": {"family_id": 0, "family": ""}})
             self._database.functions.update_many({"family_id": family_id}, {"$set": {"family_id": 0}})
+            self._updateFamilyStats(0, family_entry.num_samples, family_entry.num_functions, family_entry.num_library_samples)
         else:
             for sample_entry in sample_entries:
                 self.deleteSample(sample_entry.sample_id)
+        # ensure we always have family_id 0 as empty family.
+        if family_id == 0:
+            self._database.families.update_one({"family_id": 0}, {"$set": {"num_samples": 0, "num_functions": 0, "num_library_samples": 0}})
+        else:
+            self._database.families.delete_one({"family_id": family_id})
         self._updateDbState()
         return True
 
@@ -402,8 +515,8 @@ class MongoDbStorage(StorageInterface):
         return sample_entries
 
     def clearStorage(self) -> None:
-        collections = ["samples", "families", "functions", "matches", "candidates", "counters"]
-        for band_id in range(self._config.STORAGE_NUM_BANDS):
+        collections = ["samples", "families", "functions", "matches", "candidates", "counters", "query_samples", "query_functions"]
+        for band_id in range(self._storage_config.STORAGE_NUM_BANDS):
             collections.append("band_%d" % band_id)
         for c in collections:
             self._database[c].drop()
@@ -417,23 +530,34 @@ class MongoDbStorage(StorageInterface):
             return SampleEntry.fromDict(report_dict)
         return None
 
-    def addSmdaReport(self, smda_report: "SmdaReport") -> Optional["SampleEntry"]:
+    def addSmdaReport(self, smda_report: "SmdaReport", isQuery=False) -> Optional["SampleEntry"]:
         sample_entry = None
-        if not self.getSampleBySha256(smda_report.sha256):
-            family_id = self.addFamily(smda_report.family)
+        if isQuery:
             sample_entry = SampleEntry(
-                smda_report, sample_id=self._useCounter("samples"), family_id=family_id
+                smda_report, sample_id=-1 * self._useCounter("query_samples"), family_id=0
             )
-            self._dbInsert("samples", sample_entry.toDict())
-            function_ids = self._useCounterBulk("functions",  smda_report.num_functions)
+            self._dbInsert("query_samples", sample_entry.toDict())
+            function_ids = self._useCounterBulk("query_functions",  smda_report.num_functions)
             function_dicts = []
             for function_id, smda_function in zip(function_ids, smda_report.getFunctions()):
-                function_dicts.append(self._getFunctionDocument(sample_entry, smda_function, function_id))
-            self._dbInsertMany("functions", function_dicts)
-            self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
-            self._updateDbState()
+                function_dicts.append(self._getFunctionDocument(sample_entry, smda_function, -1 * function_id))
+            self._dbInsertMany("query_functions", function_dicts)
         else:
-            LOGGER.warn("Sample %s already existed, skipping.", smda_report.sha256)
+            if not self.getSampleBySha256(smda_report.sha256):
+                family_id = self.addFamily(smda_report.family)
+                sample_entry = SampleEntry(
+                    smda_report, sample_id=self._useCounter("samples"), family_id=family_id
+                )
+                self._dbInsert("samples", sample_entry.toDict())
+                function_ids = self._useCounterBulk("functions",  smda_report.num_functions)
+                function_dicts = []
+                for function_id, smda_function in zip(function_ids, smda_report.getFunctions()):
+                    function_dicts.append(self._getFunctionDocument(sample_entry, smda_function, function_id))
+                self._dbInsertMany("functions", function_dicts)
+                self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
+                self._updateDbState()
+            else:
+                LOGGER.warning("Sample %s already existed, skipping.", smda_report.sha256)
         return sample_entry
 
     def importSampleEntry(self, sample_entry: "SampleEntry") -> Optional["SampleEntry"]:
@@ -444,7 +568,7 @@ class MongoDbStorage(StorageInterface):
             self._updateFamilyStats(sample_entry.family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
             self._updateDbState()
         else:
-            LOGGER.warn("Sample %s already existed, skipping.", sample_entry.sha256)
+            LOGGER.warning("Sample %s already existed, skipping.", sample_entry.sha256)
         return sample_entry
 
     def importFunctionEntry(self, function_entry: "FunctionEntry") -> Optional["FunctionEntry"]:
@@ -470,13 +594,29 @@ class MongoDbStorage(StorageInterface):
     def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
         if not self.isSampleId(sample_id):
             return None
-        function_dicts = list(self._database.functions.find({"sample_id": sample_id}, {"_id": 0}))
+        if sample_id < 0:
+            function_dicts = list(self._database.query_functions.find({"sample_id": sample_id}, {"_id": 0}))
+        else:
+            function_dicts = list(self._database.functions.find({"sample_id": sample_id}, {"_id": 0}))
         functions = []
         for f in function_dicts:
             self._decodeFunction(f)
             functions.append(FunctionEntry.fromDict(f))
         return functions
 
+    def getFunctionIdsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
+        function_ids = None
+        if not self.isSampleId(sample_id):
+            return function_ids
+        function_ids = []
+        if sample_id < 0:
+            function_dicts = list(self._database.query_functions.find({"sample_id": sample_id}, {"_id": 0, "function_id": 1}))
+        else:
+            function_dicts = list(self._database.functions.find({"sample_id": sample_id}, {"_id": 0, "function_id": 1}))
+        for f in function_dicts:
+            function_ids(f["function_ids"])
+        return function_ids
+
     def getFunctions(self, start_index: int, limit: int) -> Optional["FunctionEntry"]:
         functions = []
         for function_document in self._database.functions.find().skip(start_index).limit(limit):
@@ -488,7 +628,12 @@ class MongoDbStorage(StorageInterface):
         return self._database.samples.find().distinct("sample_id")
 
     def isSampleId(self, sample_id: int) -> bool:
-        return bool(self._database.samples.find_one({"sample_id": sample_id}))
+        is_sample_id = None
+        if sample_id < 0:
+            is_sample_id = bool(self._database.query_samples.find_one({"sample_id": sample_id}))
+        else:
+            is_sample_id = bool(self._database.samples.find_one({"sample_id": sample_id}))
+        return is_sample_id
 
     def getPicHashMatchesBySampleId(self, sample_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
         if not self.isSampleId(sample_id):
@@ -502,7 +647,7 @@ class MongoDbStorage(StorageInterface):
         return self.getPicHashMatchesByFunctionIds(function_ids)
 
     def getPicHashMatchesByFunctionId(self, function_id: int) -> Optional[Dict[int, Set[Tuple[int, int, int]]]]:
-        query_result = self._database.functions.find_one({"function_id": function_id})
+        query_result = self._database.functions.find_one({"function_id": function_id}, {"_id": 0, "_pichash": 1})
         if query_result is None or not "_pichash" in query_result:
             return None
         self._decodePichash(query_result, delete_old=False)
@@ -514,7 +659,7 @@ class MongoDbStorage(StorageInterface):
         sample_and_function_ids = set(
             map(
                 lambda x: (x["family_id"], x["sample_id"], x["function_id"]),
-                list(self._database.functions.find({"_pichash": encoded_pichash})),
+                list(self._database.functions.find({"_pichash": encoded_pichash}, {"family_id": 1, "sample_id": 1, "function_id":1, "_id": 0})),
             )
         )
 
@@ -525,6 +670,7 @@ class MongoDbStorage(StorageInterface):
         pichash_raw = self._database.functions.find({"function_id": {"$in": function_ids}}, {"_pichash": 1, "_id": 0})
         pichashes_list = list(pichash_raw)  # broken? TODO
         pichashes = {}
+        fields_to_fetch =  {"family_id": 1, "sample_id": 1, "function_id":1, "_id": 0}
         for pichash in pichashes_list:
             # TODO what happens if pichash does not exist??
             self._decodePichash(pichash, delete_old=False)
@@ -535,13 +681,18 @@ class MongoDbStorage(StorageInterface):
             pichashes[decoded_pichash] = set(
                 map(
                     lambda x: (x["family_id"], x["sample_id"], x["function_id"]),
-                    list(self._database.functions.find({"_pichash": encoded_pichash})),
+                    list(self._database.functions.find({"_pichash": encoded_pichash}, fields_to_fetch)),
                 )
             )
         return pichashes
 
     def isFunctionId(self, function_id: int) -> bool:
-        return bool(self._database.functions.find_one({"function_id": function_id}))
+        is_function_id = None
+        if function_id < 0:
+            is_function_id = bool(self._database.query_functions.find_one({"function_id": function_id}))
+        else:
+            is_function_id = bool(self._database.functions.find_one({"function_id": function_id}))
+        return is_function_id
 
     def isFamilyId(self, family_id: int) -> bool:
         return bool(self._database.families.find_one({"family_id": family_id}))
@@ -583,7 +734,7 @@ class MongoDbStorage(StorageInterface):
         self._addMinHashesToBands([minhash])
 
     def _addMinHashesToBands(self, minhashes: List["MinHash"]) -> None:
-        band_hashes = {number: {} for number in range(self._config.STORAGE_NUM_BANDS)}
+        band_hashes = {number: {} for number in range(self._storage_config.STORAGE_NUM_BANDS)}
         for minhash in minhashes:
             if minhash.hasMinHash():
                 for band_number, band_hash in self.getBandHashesForMinHash(minhash).items():
@@ -592,21 +743,29 @@ class MongoDbStorage(StorageInterface):
                     band_hashes[band_number][band_hash].append(minhash.function_id)
         self._updateBands(band_hashes)
 
-    def _updateBands(self, band_hashes: Dict[int, Dict[int, List[int]]]) -> None:
+    def _updateBands(self, band_hashes: Dict[int, Dict[int, List[int]]], method="push") -> None:
+        if method not in ["push", "pull"]:
+            raise ValueError(f"MongoDbStorage._updateBands() can only do 'push' and 'pull', not '{method}'.")
         num_band_updates = 0
         for band_number, band_data in band_hashes.items():
             band_updates = []
             for band_hash, function_ids in band_data.items():
+                update_command = None
                 if len(function_ids) >= 1:
-                    push_command = {"$push": {"function_ids": {"$each": function_ids}}}
-                band_updates.append(UpdateOne({"band_hash": band_hash}, push_command, upsert=True))
-            self._database["band_%d" % band_number].bulk_write(band_updates, ordered=False)
+                    if method == "push":
+                        update_command = {"$push": {"function_ids": {"$each": function_ids}}}
+                    else:
+                        update_command = {"$pull": {"function_ids": {"$in": function_ids}}}
+                band_updates.append(UpdateOne({"band_hash": band_hash}, update_command, upsert=True))
+            if band_updates:
+                self._database["band_%d" % band_number].bulk_write(band_updates, ordered=False)
             num_band_updates += len(band_updates)
+        return num_band_updates
 
-    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"]) -> Dict[int, Set[int]]:
+    def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"], band_matches_required=1) -> Dict[int, Set[int]]:
         candidates = {}
-        target_band_hashes_per_band = {band_number: set() for band_number in range(self._config.STORAGE_NUM_BANDS)}
-        band_hash_to_function_ids = {band_number: {} for band_number in range(self._config.STORAGE_NUM_BANDS)}
+        target_band_hashes_per_band = {band_number: set() for band_number in range(self._storage_config.STORAGE_NUM_BANDS)}
+        band_hash_to_function_ids = {band_number: {} for band_number in range(self._storage_config.STORAGE_NUM_BANDS)}
         for function_id, minhash in function_id_to_minhash.items():
             if not minhash.hasMinHash():
                 continue
@@ -623,25 +782,45 @@ class MongoDbStorage(StorageInterface):
                 reference_function_ids = band_hash_to_function_ids[band_number][hit["band_hash"]]
                 for function_id in reference_function_ids:
                     if function_id not in candidates:
-                        candidates[function_id] = set()
-                    candidates[function_id].update(hit["function_ids"])
-        return candidates
+                        candidates[function_id] = {}
+                    for hit_function_id in hit["function_ids"]:
+                        if hit_function_id not in candidates[function_id]:
+                            candidates[function_id][hit_function_id] = 0
+                        candidates[function_id][hit_function_id] += 1
+        # reduce candidates based on banding requirements
+        valid_candidates = {}
+        for function_id, hit_counters in candidates.items():
+            for other_id, count in hit_counters.items():
+                if count >= band_matches_required:
+                    if function_id not in valid_candidates:
+                        valid_candidates[function_id] = set()
+                    valid_candidates[function_id].add(other_id)
+        return valid_candidates
 
-    def getCandidatesForMinHash(self, minhash: "MinHash") -> Set[int]:
+    def getCandidatesForMinHash(self, minhash: "MinHash", band_matches_required=1) -> Set[int]:
         if not minhash.hasMinHash():
             return
-        candidates = set([])
+        candidates = {}
         band_hashes = self.getBandHashesForMinHash(minhash)
         for band_number, band_hash in sorted(band_hashes.items()):
             band_hash_query = {"band_hash": band_hash}
             band_document = self._database["band_%d" % band_number].find_one(band_hash_query)
             if band_document:
-                candidates.update(band_document["function_ids"])
-        return candidates
+                for function_id in band_document["function_ids"]:
+                    if function_id not in candidates:
+                        candidates[function_id] = 0
+                    candidates[function_id] += 1
+        # reduce candidates based on banding requirements
+        valid_candidates = set([])
+        for function_id, hit_count in candidates.items():
+            if hit_count >= band_matches_required:
+                valid_candidates.add(function_id)
+        return valid_candidates
 
     def _getCacheDataForFunctionIds(self, function_ids: List[int]) -> Dict:
         cache_data = {}
         sample_ids = {}
+        sample_to_func_ids = {}
         minhashes = {}
         # process this in batches as the number of function_ids can be exceedingly large, pushing beyond Mongo's 16M limit
         for sliced_ids in zip_longest(*[iter(function_ids)]*500000):
@@ -654,8 +833,12 @@ class MongoDbStorage(StorageInterface):
                 minhash = bytes.fromhex(function_document["minhash"])
                 minhashes[function_id] = minhash
                 sample_ids[function_id] = sample_id
+                if sample_id not in sample_to_func_ids:
+                    sample_to_func_ids[sample_id] = set()
+                sample_to_func_ids[sample_id].add(function_id)
         cache_data["func_id_to_minhash"] = minhashes
         cache_data["func_id_to_sample_id"] = sample_ids
+        cache_data["sample_id_to_func_ids"] = sample_to_func_ids
         return cache_data
 
     def deleteXcfgForSampleId(self, sample_id: int) -> None:
@@ -764,14 +947,20 @@ class MongoDbStorage(StorageInterface):
         field_selection = {"_id": 0}
         if not with_xcfg:
             field_selection["_xcfg"] = 0
-        function_document = self._database.functions.find_one({"function_id": function_id}, field_selection)
+        if function_id < 0:
+            function_document = self._database.query_functions.find_one({"function_id": function_id}, field_selection)
+        else:
+            function_document = self._database.functions.find_one({"function_id": function_id}, field_selection)
         if not function_document:
             return None
         self._decodeFunction(function_document)
         return FunctionEntry.fromDict(function_document)
 
     def getSampleById(self, sample_id: int) -> Optional["SampleEntry"]:
-        sample_document = self._database.samples.find_one({"sample_id": sample_id}, {"_id": 0})
+        if sample_id < 0:
+            sample_document = self._database.query_samples.find_one({"sample_id": sample_id}, {"_id": 0})
+        else:
+            sample_document = self._database.samples.find_one({"sample_id": sample_id}, {"_id": 0})
         if not sample_document:
             return None
         return SampleEntry.fromDict(sample_document)
@@ -787,25 +976,42 @@ class MongoDbStorage(StorageInterface):
             "db_state": self._getDbState(),
             "num_families": self._database.families.count_documents(filter={}),
             "num_samples": self._database.samples.count_documents(filter={}),
+            "num_query_samples": self._database.query_samples.count_documents(filter={}),
+            "num_query_functions": self._database.query_functions.count_documents(filter={}),
             "num_functions": self._database.functions.count_documents(filter={}),
-            "num_bands": self._config.STORAGE_NUM_BANDS,
+            "num_bands": self._storage_config.STORAGE_NUM_BANDS,
             "num_pichashes": num_unique_pichashes,
         }
         return stats
 
-    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None) -> List["FunctionEntry"]:
+    def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None, only_function_ids=False) -> List["FunctionEntry"]:
         unhashed_functions = []
         search_query = {}
         if function_ids is not None:
             search_query = {"function_id": {"$in": list(function_ids)}}
         for function_document in self._database.functions.find(search_query, {"_id": 0}):
-            self._decodeFunction(function_document)
-            if not function_document["minhash"]:
-                unhashed_functions.append(FunctionEntry.fromDict(function_document))
+            if function_document["minhash"] == "":
+                if only_function_ids:
+                    unhashed_functions.append(function_document["function_id"])
+                else:
+                    self._decodeFunction(function_document)
+                    unhashed_functions.append(FunctionEntry.fromDict(function_document))
         return unhashed_functions
 
-    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None) -> Dict:
+    def getUniqueBlocks(self, sample_ids: Optional[List[int]] = None, progress_reporter=None) -> Dict:
         # query once to get all blocks from the functions of our samples
+        block_statistics = {
+            "by_sample_id": {
+                sample_id: {
+                    "sample_id": sample_id,
+                    "total_blocks": 0,
+                    "characteristic_blocks": 0,
+                    "unique_blocks": 0
+                } for sample_id in sample_ids
+            },
+            "unique_blocks_overall": 0,
+            "num_samples": len(sample_ids)
+        }
         candidate_picblockhashes = {}
         for entry in self._database.functions.find({"sample_id": {"$in": sample_ids}, "_picblockhashes": {"$exists": True, "$ne": [] }}, {"function_id": 1, "sample_id": 1, "_picblockhashes": 1, "_id": 0}):
             sample_id = entry["sample_id"]
@@ -819,21 +1025,50 @@ class MongoDbStorage(StorageInterface):
                         "sample_id": sample_id,
                         "offset": block_entry["offset"],
                         "instructions": [],
-                        "escaped_sequence": ""
+                        "escaped_sequence": "",
+                        "score": 0
                     }
                 candidate_picblockhashes[block_hash]["samples"].add(sample_id)
+        # update statistics based on candidates
+        for picblockhash, entry in candidate_picblockhashes.items():
+            for sample_id in entry["samples"]:
+                block_statistics["by_sample_id"][sample_id]["total_blocks"] += 1
         LOGGER.info(f"Found {len(candidate_picblockhashes)} candidate picblock hashes")
+        if progress_reporter is not None:
+            progress_reporter.set_total(self._database.functions.count_documents(filter={}))
         # remove those that are not unique
         for entry in self._database.functions.find({"_picblockhashes": {"$exists": True, "$ne": [] }}, {"sample_id": 1, "_picblockhashes": 1, "_id": 0}):
+            if progress_reporter is not None:
+                progress_reporter.step()
             sample_id = entry["sample_id"]
             if sample_id not in sample_ids:
                 for block_entry in entry["_picblockhashes"]:
                     candidate_picblockhashes.pop(block_entry["hash"], None)
+        # update statistics again after having reduced to results
+        for picblockhash, entry in candidate_picblockhashes.items():
+            if len(entry["samples"]) == 1:
+                single_sample_id = list(entry["samples"])[0]
+                block_statistics["by_sample_id"][single_sample_id]["unique_blocks"] += 1
+            for sample_id in entry["samples"]:
+                block_statistics["by_sample_id"][sample_id]["characteristic_blocks"] += 1
+        block_statistics["unique_blocks_overall"] = len(candidate_picblockhashes)
         LOGGER.info(f"Reduced to {len(candidate_picblockhashes)} unique picblock hashes")
+        # we are basically finished when we reached this step, so set progress to 100%
+        if progress_reporter is not None:
+            progress_reporter.set_total(1)
+            progress_reporter.step()
         # iterate over candidates by function_id and extract instructions
         function_id_to_block_offsets = {}
         for picblockhash, entry in candidate_picblockhashes.items():
             candidate_picblockhashes[picblockhash]["samples"] = sorted(list(entry["samples"]))
+            # we calculate the score for this block as 80% of how well it covers the samples and 20% how far its size is away from an "ideal" signature block
+            sample_score = 100.0 * len(entry["samples"]) / len(sample_ids)
+            length_score = 100.0
+            if entry["length"] < 7:
+                length_score = 100.0 - (100.0 * (7 - entry["length"]) / 7)
+            elif entry["length"] > 10:
+                length_score = 100.0 * (1 / (entry["length"] - 10))
+            candidate_picblockhashes[picblockhash]["score"] = 0.8 * sample_score + 0.2 * length_score
             if entry["function_id"] not in function_id_to_block_offsets:
                 function_id_to_block_offsets[entry["function_id"]] = []
             function_id_to_block_offsets[entry["function_id"]].append((entry["offset"], picblockhash))
@@ -843,7 +1078,7 @@ class MongoDbStorage(StorageInterface):
             for block_offset, picblockhash in function_id_to_block_offsets[function_id]:
                 candidate_picblockhashes[picblockhash]["instructions"] = entry["xcfg"]["blocks"][str(block_offset)]
         LOGGER.info(f"Instructions for {len(candidate_picblockhashes)} blocks extracted.")
-        return candidate_picblockhashes
+        return {"statistics": block_statistics, "unique_blocks": candidate_picblockhashes}
 
 
     ##### helpers for search ######
